/*
    Small library for reading and writing audio.

    This library forks an audio task and communicates with it
    via a shared memory segment.

    All what this library does can be done in principal
    with ordinary read/write calls on the sound device.

    The Linux audio driver uses so small buffers, however,
    that overruns/underruns are unavoidable in many cases.

    Copyright (C) 1999 Rainer Johanni <Rainer@Johanni.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <linux/soundcard.h>
#include <sys/resource.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/ioctl.h>


#define DEBUG(x) 

#ifdef USE_ALSA
#include <sys/asoundlib.h>
#endif

/* The shared memory things */

#include <sys/ipc.h>
#include <sys/shm.h>

static int pid; /* pid of child */
static int shm_seg;



#define N_SHM_BUFFS 256 /* Number of buffers, must be a power of 2 */
#define SHM_BUFF_MASK (N_SHM_BUFFS-1)
#define BUFFSIZE (65536)

struct shm_buff_s
{
   volatile unsigned char audio_data[N_SHM_BUFFS][BUFFSIZE];
   volatile int used_flag[N_SHM_BUFFS];
   volatile int exit_flag;
   volatile int audio_status;
   volatile int audio_sync;
} *shmemptr;

static int audio_buffer_size = BUFFSIZE;    /* The buffer size actually used */

/* The parameters for audio capture/playback */

static int initialized=0;
static int audio_capt;           /* Flag for capture/playback */
static int stereo;               /* 0: capture mono, 1: capture stereo */
static int audio_size;           /* size of an audio sample: 8 or 16 bits */
static int audio_rate;           /* sampling rate for audio */

/* Buffer counter */

static int n_audio;

/* Bookkeeping of the write buffers */

static char audio_left_buf[BUFFSIZE];  
static int audio_bytes_left=0;   /* Number of bytes in audio_left_buf */

/* Forward declarations: */

void do_audio();

/*
 * audio_init: Initialize audio system.
 *
 *      a_read     0: User is going to write (output) audio
 *                 1: User is going to read (input) audio
 *      a_stereo   0: mono, 1: stereo
 *      a_size     size of an audio sample: 8 or 16 bits
 *      a_rate     sampling rate for audio
 *
 *      returns 0 for success, -1 for failure
 *
 */

int audio_init(int a_read, int a_stereo, int a_size, int a_rate)
{
   int tmp, i;

   /* Check if the audio task is allready initialized */

   if(initialized)
   {
      fprintf(stderr,"audio_init: Audio task is allready initialized\n");
      return -1;
   }

   /* Checks of parameters */

   if (a_size != 8 && a_size != 16)
   {
      fprintf(stderr,"audio_init: audio size = %d, must be 8 or 16\n",a_size);
      return -1;
   }

   /* Copy our parameters into static space */

   audio_capt = a_read;
   stereo     = a_stereo;
   audio_size = a_size;
   audio_rate = a_rate;

   /* Buffer counter */

   n_audio = 0;

   if (audio_capt)
   {
      /*
       * Calculate bytes/second of the audio stream
       * Set audio buffer size accordingly
       */

      tmp = audio_rate;
      if (stereo)         tmp *= 2;
      if (audio_size==16) tmp *= 2;

      audio_buffer_size = 65536;
      if(tmp<88200) audio_buffer_size = 32768;
      if(tmp<44100) audio_buffer_size = 16384;
   }
   else
   {
      /* For writing we allways use the maximum buffer size */

      audio_buffer_size = BUFFSIZE;
   }

   /* Allocate shared memory segment */

   shm_seg = shmget(IPC_PRIVATE, sizeof(struct shm_buff_s), IPC_CREAT | 0777);
   if(shm_seg < 0) { perror("shmget"); return -1; }
// fprintf(stderr,"shm_seg = %d\n",shm_seg);

   /* attach the segment and get its address */

   shmemptr = (struct shm_buff_s *) shmat(shm_seg,0,0);
   if(shmemptr < 0) { perror("shmat"); return -1; }

   /* mark the segment as destroyed, it will be removed after
      the last process which had attached it is gone */

   if( shmctl( shm_seg, IPC_RMID, (struct shmid_ds *)0 ) == -1 )
   {
      perror("remove shared memory segment");
      return -1;
   }

   /* set the flags in the shared memory */

   for(i=0;i<N_SHM_BUFFS;i++) shmemptr->used_flag[i] = 0;
   shmemptr->exit_flag    = 0;
   shmemptr->audio_status = 0;
   shmemptr->audio_sync   = 0;

   /* do the fork */

   pid = fork();
   if(pid<0)
   {
      perror("fork");
      return -1;
   }

   /* the child goes into the audio task */

   if (pid==0)
   {
      /* The audio buffers in Linux are ridiculosly small,
         therefore the audio task must have a high priority */

      if (setpriority(PRIO_PROCESS, getpid(), -20) < 0)
         fprintf(stderr,"Can't raise priority - Am I superuser ?\n");

      do_audio();
      exit(0);
   }

   /* the parent returns */

   initialized = 1;
   return 0;
}

/*
 * audio_shutdown: Shutdown audio system
 *
 * It is important that this routine is called whenever the host
 * program finished, or else there will be the audio task
 * left over, having the sound device still opened and preventing
 * other programs from using sound.
 *
 */

void audio_shutdown()
{
   if(!initialized) return;

   /* show the child we want to exit */

   shmemptr->exit_flag = 1;

   waitpid(pid,0,0);

   initialized = 0;
}

/*
 * audio_start: Actually trigger the start of audio after all
 *              initializations have been done.
 *
 *      returns 0 for success, -1 for failure
 */

int audio_start()
{
   if(!initialized)
   {
      fprintf(stderr,"audio_start: audio not initialized\n");
      return -1;
   }

   /* signal the audio task that we are going to start */

   shmemptr->audio_sync = 1;

   /* wait for the audio task to become ready */

   while (shmemptr->audio_status==0) usleep(10000);
   if(shmemptr->audio_status == -1) return -1;

   return 0;
}

/*
 * swpcpy: like memcpy, but bytes are swapped during copy
 */

void swpcpy(char *dst, char *src, int num)
{
   int i;

   num &= ~1; /* Safety first */

   for(i=0;i<num;i+=2)
   {
      dst[i  ] = src[i+1];
      dst[i+1] = src[i  ];
   }
}

/*
 * audio_read: Get audio data, if available
 *             Behaves like a nonblocking read
 *
 *    buf      Buffer where to copy the data
 *    size     Size of the buffer
 *    swap     Flag if to swap the endian-ness of 16 bit data
 *
 *    returns  the number of bytes in the buffer,
 *             0 if nothing available
 *            -1 if an error occured
 *
 */

int audio_read(char *buf, int size, int swap)
{
   if(!initialized)
   {
      fprintf(stderr,"audio_read: audio not initialized\n");
      return -1;
   }

   /* Is audio task still ok ? */

   if(shmemptr->audio_status < 0)
   {
      fprintf(stderr,"audio_read: Audio task died\n");
      return -1;
   }

   if(!audio_capt)
   {
      fprintf(stderr,"audio_read: Audio is in write mode\n");
      return -1;
   }

   if(size<audio_buffer_size)
   {
      fprintf(stderr,"audio_read: Buffer size too small\n");
      return -1;
   }

   /* Check if a new audio sample is ready */

   if(shmemptr->used_flag[n_audio&SHM_BUFF_MASK])
   {
      /* Got an audio sample, copy it to the output buffer */

      if(swap && audio_size==16)
         swpcpy(buf,(void*)shmemptr->audio_data[n_audio&SHM_BUFF_MASK],audio_buffer_size);
      else
         memcpy(buf,(void*)shmemptr->audio_data[n_audio&SHM_BUFF_MASK],audio_buffer_size);
      
      /* reset used_flag, increment n-audio */

      shmemptr->used_flag[n_audio&SHM_BUFF_MASK] = 0;
      n_audio++;
      return audio_buffer_size;
   }

   return 0;
}

/*
 * audio_write: Buffer audio data for output
 *              Behaves like a nonblocking write
 *
 *    buf       Buffer with audio data
 *    size      Size of the buffer
 *    swap      Flag if to swap the endian-ness of 16 bit data
 *
 *    returns   the number of bytes actually written
 *              -1 if an error occured
 *
 *              If the number of bytes actually written is smaller
 *              than size, the audio ringbuffer is completely filled
 *
 */

int audio_write(char *buf, int size, int swap)
{
   int nb;

   if(!initialized)
   {
      fprintf(stderr,"audio_write: audio not initialized\n");
      return -1;
   }

   /* Is audio task still ok ? */

   if(shmemptr->audio_status < 0)
   {
      fprintf(stderr,"audio_write: Audio task died\n");
      return -1;
   }

   if(audio_capt)
   {
      fprintf(stderr,"audio_write: Audio is in read mode\n");
      return -1;
   }

   /* If the number of bytes we got isn't big enough to fill
      the next buffer, copy buf into audio_left_buf */

   if (audio_bytes_left+size < BUFFSIZE)
   {
      memcpy(audio_left_buf+audio_bytes_left,buf,size);
      audio_bytes_left += size;
      return size;
   }

   nb = 0;

   /* if audio_left_buf contains something, output that first */

   if (audio_bytes_left)
   {
      memcpy(audio_left_buf+audio_bytes_left,buf,BUFFSIZE-audio_bytes_left);

      if(shmemptr->used_flag[n_audio&SHM_BUFF_MASK]) return 0;

      if(swap && audio_size==16)
         swpcpy((void*)shmemptr->audio_data[n_audio&SHM_BUFF_MASK],audio_left_buf,BUFFSIZE);
      else
         memcpy((void*)shmemptr->audio_data[n_audio&SHM_BUFF_MASK],audio_left_buf,BUFFSIZE);

      shmemptr->used_flag[n_audio&SHM_BUFF_MASK] = 1;

      nb = BUFFSIZE-audio_bytes_left;
      audio_bytes_left = 0;
      n_audio++;
   }

   /* copy directly to the shmem buffers */

   while(size-nb >= BUFFSIZE)
   {
      if(shmemptr->used_flag[n_audio&SHM_BUFF_MASK]) return nb;

      if(swap && audio_size==16)
         swpcpy((void*)shmemptr->audio_data[n_audio&SHM_BUFF_MASK],buf+nb,BUFFSIZE);
      else
         memcpy((void*)shmemptr->audio_data[n_audio&SHM_BUFF_MASK],buf+nb,BUFFSIZE);

      shmemptr->used_flag[n_audio&SHM_BUFF_MASK] = 1;

      nb += BUFFSIZE;
      n_audio++;
   }

   /* copy the remainder into audio_left_buf */

   if(nb<size)
   {
      audio_bytes_left = size-nb;
      memcpy(audio_left_buf,buf+nb,audio_bytes_left);
   }

   return size;
}

/*
 * The audio task
 */

/* Timeout (in seconds) for playing audio:
 * If all buffers are played and the calling program doesn't
 * deliver anything during that time intervall, the audio task will exit
 */

#define WRITE_TIMEOUT 10

#ifndef USE_ALSA
void do_audio()
{

   int fd, tmp, nbout, nbuff, ret, caps, afmt, frag, i, last_blocks, wait_time;

   unsigned char *buf;
   fd_set selectset;

   struct count_info count;
   struct audio_buf_info info;


/*
 * Fragment size and max possible number of frags
 */ 

   switch (audio_buffer_size)
   {
      case 65536: frag = 0x7fff000d; break;
      case 32768: frag = 0x7fff000c; break;
      case 16384: frag = 0x7fff000b; break;
      default:
         fprintf(stderr,"Audio internal error\n");
         shmemptr->audio_status = -1;
         exit(1);
   }
   /* if somebody plays with BUFFSIZE without knowing what he does ... */
   if (audio_buffer_size>BUFFSIZE)
   {
      fprintf(stderr,"Audio internal error\n");
      shmemptr->audio_status = -1;
      exit(1);
   }

/*
 * Open Audio device, set number of frags wanted
 */

   if(audio_capt)
      fd=open("/dev/dsp", O_RDONLY, 0);
   else
      fd=open("/dev/dsp", O_RDWR,   0);

   if (fd<0)
   {
      perror("Opening /dev/dsp");
      shmemptr->audio_status = -1;
      exit(1);
   }

   if (ioctl(fd, SNDCTL_DSP_SETFRAGMENT, &frag)<0)
   {
      perror("SETFRAGMENT");
      shmemptr->audio_status = -1;
      exit(1);
   }

/*
 * Setup sampling parameters.
 */

   afmt = (audio_size==16) ? AFMT_S16_LE : AFMT_U8;
   tmp = afmt;
   ret = ioctl(fd, SNDCTL_DSP_SETFMT, &tmp);
   if(ret<0 || tmp!=afmt)
   {
      fprintf(stderr,"Unable to set sound format\n");
      shmemptr->audio_status = -1;
      exit(1);
   }

   tmp = stereo; /* 0=mono, 1=stereo */
   ret = ioctl(fd, SNDCTL_DSP_STEREO, &tmp);
   if(ret<0 || tmp!=stereo)
   {
      fprintf(stderr,"Unable to set %s\n", stereo ? "STEREO" : "MONO");
      shmemptr->audio_status = -1;
      exit(1);
   }

   tmp = audio_rate;
   ret = ioctl(fd, SNDCTL_DSP_SPEED, &tmp);
   if(ret<0 || tmp!=audio_rate)
   {
      fprintf(stderr,"Unable to set sound rate %d\n",audio_rate);
      shmemptr->audio_status = -1;
      exit(1);
   }

/*
 * Check that the device has capability to do mmap and trigger
 */

   if (ioctl(fd, SNDCTL_DSP_GETCAPS, &caps)<0)
   {
      fprintf(stderr,"Cant get audio device capabilities\n");
      shmemptr->audio_status = -1;
      exit(1);
   }
   if (!(caps & DSP_CAP_TRIGGER) || !(caps & DSP_CAP_MMAP))
   {
      fprintf(stderr,"Soundcard cant do mmap or trigger\n");
      shmemptr->audio_status = -1;
      exit(1);
   }

/*
 * Get the size of the input/output buffer and do the mmap
 */

   if (audio_capt)
      ret = ioctl(fd, SNDCTL_DSP_GETISPACE, &info);
   else
      ret = ioctl(fd, SNDCTL_DSP_GETOSPACE, &info);

   if (ret<0)
   {
      perror("SNDCTL_DSP_GET[IO]SPACE");
      shmemptr->audio_status = -1;
      exit(1);
   }

   if (info.fragsize != audio_buffer_size)
   {
      fprintf(stderr,"Soundcard fragment size is %d --- must be %d\n",
              info.fragsize,audio_buffer_size);
      shmemptr->audio_status = -1;
      exit(1);
   }

/*
 * Normally we should get at least 8 fragments (if we use 8KB buffers)
 * or even more if we use 4KB od 2 KB buffers
 * We consider 4 fragments as the absolut minimum here!
 */

   if (info.fragstotal < 4)
   {
      fprintf(stderr,"Could not get enough audio buffers (only %d KB)\n",
              info.fragstotal*info.fragsize/1024);
      shmemptr->audio_status = -1;
      exit(1);
   }

   tmp = info.fragstotal*info.fragsize;
   if (audio_capt)
      buf=mmap(NULL, tmp, PROT_READ , MAP_SHARED, fd, 0);
   else
      buf=mmap(NULL, tmp, PROT_WRITE, MAP_SHARED, fd, 0);

   if (buf==MAP_FAILED)
   {
      perror("mmap of audio buffer");
      shmemptr->audio_status = -1;
      exit(1);
   }


/*
 * Put device into hold
 */

   tmp = 0;
   if(ioctl(fd, SNDCTL_DSP_SETTRIGGER, &tmp)<0)
   {
      perror("SNDCTL_DSP_SETTRIGGER");
      shmemptr->audio_status = -1;
      exit(1);
   }

/*
 * Wait until the parent tells us it wants to start
 */

   while (!shmemptr->audio_sync && !shmemptr->exit_flag) usleep(10000);
   if(shmemptr->exit_flag) exit(0);

/*
 * nbout is the number of fragments which have been output so far
 *       (not used in audio capture)
 * nbuff is the number of buffers which have been
 *       queued/filled so far.
 *
 * If we do playback:
 * Queue up all audio samples put into the shared memory so far
 */

   nbuff = 0; /* was: nwritten */
   nbout = 0;

   if(!audio_capt)
   {
      for(i=0;i<info.fragstotal;i++)
      {
         if(!shmemptr->used_flag[nbuff&SHM_BUFF_MASK]) break;
         memcpy(buf+(nbuff%info.fragstotal)*info.fragsize,
                (void*) shmemptr->audio_data[nbuff&SHM_BUFF_MASK],
                info.fragsize);
         /* Mark the buffer as free */
         shmemptr->used_flag[nbuff&SHM_BUFF_MASK] = 0;
         nbuff++;
      }
   }


/*
 * Signal the parent that we are ready
 */

   shmemptr->audio_status = 1;


/*
 * Fire up audio device
 */

   if(audio_capt)
      tmp = PCM_ENABLE_INPUT;
   else
      tmp = PCM_ENABLE_OUTPUT;

   if(ioctl(fd, SNDCTL_DSP_SETTRIGGER, &tmp)<0)
   {
      perror("\nSNDCTL_DSP_SETTRIGGER");
      shmemptr->audio_status = -1;
      exit(1);
   }


   /* The recording/playback loop */

   last_blocks = 0;

   while(1)
   {
      /* Wait until new audio data can be read/written */

      FD_ZERO(&selectset);
      FD_SET(fd, &selectset);

      if(audio_capt)
         ret = select(fd+1, &selectset, NULL, NULL, NULL);
      else
         ret = select(fd+1, NULL, &selectset, NULL, NULL);

      if (ret<0)
      {
         perror("\nWaiting on audio with select");
         shmemptr->audio_status = -1;
         exit(1);
      }

      /* If we do play: Wait for new audio data available */

      if(!audio_capt)
      {
         wait_time = 0;
         while(!shmemptr->used_flag[nbuff&SHM_BUFF_MASK] &&
               !shmemptr->exit_flag)
         {
            wait_time += 100; /* in ms */
            if(wait_time > WRITE_TIMEOUT*1000)
            {
               fprintf(stderr,"\nAudio playback timeout\n");
               shmemptr->audio_status = -1;
               exit(1);
            }
            usleep(100000); /* sleep 100 ms */
         }
      }

      /* Get the status of the sound buffer */

      if(audio_capt)
         ret = ioctl(fd, SNDCTL_DSP_GETIPTR, &count);
      else
         ret = ioctl(fd, SNDCTL_DSP_GETOPTR, &count);

      if (ret<0)
      {
         perror("\nSNDCTL_DSP_GET[IO]PTR");
         shmemptr->audio_status = -1;
         exit(1);
      }

      if(audio_capt)
      {
         /* if exit_flag is set, exit immediatly */

         if(shmemptr->exit_flag)
         {
            shmemptr->audio_status = -1;
            exit(0);
         }

         /* output a warning if a buffer overrun occured during read */

         if(count.blocks>info.fragstotal)
            fprintf(stderr,"\n--------------- Audio recording overrun\n");
         else if(count.blocks+last_blocks>info.fragstotal)
            fprintf(stderr,"\n--------------- Possible audio recording overrun\n");
         last_blocks = count.blocks;

         /* copy the ready buffers to our audio ring buffer */

         for(;count.blocks;count.blocks--)
         {

            /* Check if buffer nbuff in the ring buffer is free */

            if(shmemptr->used_flag[nbuff&SHM_BUFF_MASK])
            {
               fprintf(stderr,"\nAudio ring buffer overflow\n");
               shmemptr->audio_status = -1;
               exit(1);
            }

            memcpy((void*) shmemptr->audio_data[nbuff&SHM_BUFF_MASK],
                   buf+(nbuff%info.fragstotal)*info.fragsize,
                   info.fragsize);

            shmemptr->used_flag[nbuff&SHM_BUFF_MASK] = 1;

            nbuff++;
         }
      }
      else
      {
         /* Update the number of frags(=buffers) already output */

         nbout += count.blocks;

         /* We have a buffer underrun during write if nbout>=nbuff.
            As long as the buffers are delivered later, there will
            be some audible artifacts, but we stay in sync
            - we don't output a warning */

         /* If exit_flag is set and all buffers are played, exit */

         if(shmemptr->exit_flag && nbout >= nbuff)
         {
            shmemptr->audio_status = -1;
            exit(0);
         }

         /* Fill into the soundcard memory as many buffers
            as fit and are available */

         while(nbuff-nbout < info.fragstotal)
         {
            if(!shmemptr->used_flag[nbuff&SHM_BUFF_MASK]) break;

            memcpy(buf+(nbuff%info.fragstotal)*info.fragsize,
                   (void*) shmemptr->audio_data[nbuff&SHM_BUFF_MASK],
                   info.fragsize);

            /* Mark the buffer as free */
            shmemptr->used_flag[nbuff&SHM_BUFF_MASK] = 0;

            nbuff++;
         }
      }
   }
}



#else
/* do_audio (ALSA version) 
 * The audio playback/record task for ALSA.
 * returns: nothing (paralell task)
 */

void do_audio()
{
   int nbuff, err, wait_time;
   int card = 0, device = 0; 

   struct snd_pcm *handle; /* like a file descriptor */
   struct snd_pcm_playback_params play_fragparam; /* fragmentation buffer setup */
   struct snd_pcm_playback_status play_info; /* status info record */
   struct snd_pcm_capture_params record_fragparam;
   struct snd_pcm_capture_status record_info;
   struct snd_pcm_format format; /* sets the record/playback format */

   if (audio_capt)
     {
       if ((err = snd_pcm_open(&handle, card, device, 
			       SND_PCM_OPEN_CAPTURE)) < 0) 
	 {  
	   fprintf(stderr, "ALSA open for record failed: %s\n", snd_strerror( err ));  
	   shmemptr->audio_status = -1; exit(1);
	 } 
     }
   else
     {
       if ((err = snd_pcm_open(&handle, card, device, 
			       SND_PCM_OPEN_PLAYBACK)) < 0) 
	 {  
	   fprintf(stderr, "ALSA open for playback failed: %s\n", snd_strerror( err ));  
	   shmemptr->audio_status = -1; exit(1);
	 } 
     }

   bzero(&format, sizeof(format)); 
   format.format = (audio_size == 16) ?  SND_PCM_SFMT_S16_LE : SND_PCM_SFMT_U8; 
   format.rate = audio_rate; 
   format.channels = stereo ? 2 : 1; 

   if (audio_capt)
     {
       if ((err = snd_pcm_capture_format(handle, &format)) < 0) 
	 { 
	   fprintf(stderr, "format setup for record failed: %s\n", snd_strerror( err ));  
	   snd_pcm_close(handle); 
	   shmemptr->audio_status = -1; exit(1);
	 } 
     }
   else
     {
       if ((err = snd_pcm_playback_format(handle, &format)) < 0) 
	 { 
	   fprintf(stderr, "format setup for playback failed: %s\n", snd_strerror( err ));  
	   snd_pcm_close(handle); 
	   shmemptr->audio_status = -1; exit(1);
	 } 
     }


   /* if somebody plays with BUFFSIZE without knowing what he does ... */
   if (audio_buffer_size > BUFFSIZE)
     {
       fprintf(stderr,"Audio internal error\n");
       shmemptr->audio_status = -1; exit(1);
     }

   if (audio_capt)
     {
       bzero(&record_fragparam, sizeof(record_fragparam)); 
       record_fragparam.fragment_size = audio_buffer_size;
       record_fragparam.fragments_min = 1;
       err = snd_pcm_capture_params(handle, &record_fragparam); 
     }
   else
     {
       bzero(&play_fragparam, sizeof(play_fragparam)); 
       play_fragparam.fragment_size = audio_buffer_size;
       play_fragparam.fragments_max = -1;
       play_fragparam.fragments_room = 1;
       err = snd_pcm_playback_params(handle, &play_fragparam); 
     }
   if (err < 0)
    { 
      fprintf(stderr, "%s parameter setup failed: %s\n", 
	      (audio_capt) ? "recording" : "playback", snd_strerror( err ));        
      snd_pcm_close(handle); 
      shmemptr->audio_status = -1; exit(1);
    } 

   /* Pause the playback */
   /* I have no clue about how to pause recording :-/ */
   if (!audio_capt)
       snd_pcm_playback_pause(handle, 1);

/* Wait until the parent tells us it wants to start */

   while (!shmemptr->audio_sync && !shmemptr->exit_flag) usleep(10000);
   if(shmemptr->exit_flag) exit(0);

   /* "Play" (=queue) all audio samples put into the shared memory so far */
   nbuff = 0;

   if (!audio_capt)
     while (shmemptr->used_flag[nbuff & SHM_BUFF_MASK]) 
       {
	 snd_pcm_write(handle, (void*) shmemptr->audio_data[nbuff & SHM_BUFF_MASK],
		       play_info.fragment_size);
	 /* Mark the buffer as free */
	 shmemptr->used_flag[nbuff & SHM_BUFF_MASK] = 0;
	 nbuff++;
       }

   if (!audio_capt)
     snd_pcm_playback_pause(handle, 0);

/* Signal the parent that we are ready */

   shmemptr->audio_status = 1;

   /* The record/playback loop */
   while (1)
   {
      /* If we do play: Wait for new audio data available */
      if(!audio_capt)
      {
         wait_time = 0;
         while(!shmemptr->used_flag[nbuff&SHM_BUFF_MASK] &&
               !shmemptr->exit_flag)
         {
            wait_time += 100; /* in ms */
            if(wait_time > WRITE_TIMEOUT*1000)
            {
               fprintf(stderr,"\nAudio playback timeout\n");
	       snd_pcm_close(handle); 
               shmemptr->audio_status = -1; exit(1);
            }
            usleep(100000); /* sleep 100 ms */
         }
      }

      if (audio_capt)
      { 
	if ((err = snd_pcm_capture_status(handle, &record_info)) < 0) 
	  { 
	    fprintf(stderr, "Could not retrieve recording status: %s\n", 
		    snd_strerror( err ));      
	    snd_pcm_close(handle); 
	    shmemptr->audio_status = -1; exit(1);
	  }; 
      
	DEBUG(
	  { printf("Record status info from ALSA:\n"
		   "%d fragments with %d bytes size.\n"
		   "%d bytes readable without blocking. %d overruns.\n",
		   record_info.fragments, record_info.fragment_size,
		   record_info.count, record_info.overrun);
	  })
      }
      else
	{
	  if ((err = snd_pcm_playback_status(handle, &play_info)) < 0) 
	    { 
	      fprintf(stderr, "Could not retrieve playback status: %s\n", 
		      snd_strerror( err ));      
	      snd_pcm_close(handle); 
	      shmemptr->audio_status = -1; exit(1);
	    } 
	  
	  DEBUG(
	    { printf("Playback status info from ALSA:\n"
		     "%d fragments with %d bytes size.\n"
		     "%d bytes writable without blocking. %d bytes in queue.\n",
		     play_info.fragments, play_info.fragment_size,
		     play_info.count, play_info.queue);
	    })

	}

      if(audio_capt)
      {
         /* if exit_flag is set, exit immediatly */
         if(shmemptr->exit_flag)
         {
	   snd_pcm_close(handle);
	   shmemptr->audio_status = -1; exit(0);
         }

         /* output a warning if a buffer overrun occured during read */
         if (record_info.overrun)
            fprintf(stderr,"\n*** %d Audio recording overrun(s) !!\n", 
		    record_info.overrun);

	 /* Check if buffer nbuff in the ring buffer is free */
	 if(shmemptr->used_flag[nbuff&SHM_BUFF_MASK])
	   {
	     fprintf(stderr,"\nAudio ring buffer overflow\n");
	     shmemptr->audio_status = -1;
	     exit(1);
	   }

	 snd_pcm_read (handle, 
		    (void*) shmemptr->audio_data[nbuff & SHM_BUFF_MASK], 
		       record_info.fragment_size);
	 shmemptr->used_flag[nbuff & SHM_BUFF_MASK] = 1;
	 
	 nbuff++;
      }
      else /* this is playback */
      {
         /* We can have a buffer underrun if no new samples are delivered in time.
            As long as the buffers are delivered later, there will
            be some audible artifacts, but we stay in sync
            - we don't output a warning */

         /* If exit_flag is set and all buffers are played, exit */
         if(shmemptr->exit_flag)
         {
            shmemptr->audio_status = -1;
	    snd_pcm_flush_playback(handle);
	    snd_pcm_close(handle);
            exit(0);
         }

	 /* Fill into the soundcard memory as many buffers as are available */
	 while(shmemptr->used_flag[nbuff & SHM_BUFF_MASK])
	   {
	     snd_pcm_write(handle, (void*) shmemptr->audio_data[nbuff & SHM_BUFF_MASK],
			   play_info.fragment_size);
	     /* Mark the buffer as free */
	     shmemptr->used_flag[nbuff & SHM_BUFF_MASK] = 0;
	     nbuff++;
	   }
      }
   }
}

#endif
































/* Test main() programs for reading and writing */

#if defined(TEST_READ_AUDIO) || defined(TEST_WRITE_AUDIO)

#define STEREO 0
#define A_SIZE 8
#define A_RATE 22050

#define A_TIME 20 /* Time for read in seconds */

#define BFSZ 20000 /* write buffer size */

#endif


#ifdef TEST_READ_AUDIO

/* Test program for reading audio */

void main()
{
   int nb, nr, n;
   char buf[65536];

   nb = A_RATE * A_SIZE/8 * (STEREO ? 2 : 1) * A_TIME; /* bytes to read */
   n  = 0; /* bytes transferred */

   if(audio_init(1,STEREO,A_SIZE,A_RATE)) {
      fprintf(stderr,"audio_init failed\n");
      exit(1);
   }

   if(audio_start()) {
      fprintf(stderr,"audio_start failed\n");
      exit(1);
   }

   while(n<nb)
   {
      nr = audio_read(buf,sizeof(buf),0);
      if(nr==0)
      {
         usleep(100000);
         continue;
      }
      if(nr<0) {
         fprintf(stderr,"read failed\n");
         audio_shutdown();
         exit(1);
      }
      if(fwrite(buf,nr,1,stdout)!=1) {
         fprintf(stderr,"Write to stdout failed\n");
         audio_shutdown();
         exit(1);
      }
      n += nr;
   }

   audio_shutdown();
}
#endif





#ifdef TEST_WRITE_AUDIO

/* Test program for writing audio */

void main()
{
   int nr, nw, n;
   char buf[BFSZ];

   /* Read the first buffer */

   if(fread(buf,1,BFSZ,stdin)!=BFSZ) {
      fprintf(stderr,"Read from stdin failed\n");
      exit(1);
   }

   if(audio_init(0,STEREO,A_SIZE,A_RATE)) {
      fprintf(stderr,"audio_init failed\n");
      exit(1);
   }

   /* Put first buffer into sound buffers before actually starting audio */

   if(audio_write(buf,BFSZ,0) != BFSZ) {
      fprintf(stderr,"audio_write failed\n");
      audio_shutdown();
      exit(1);
   }

   if(audio_start()) 
     {
       fprintf(stderr,"audio_start failed\n");
       audio_shutdown();
      exit(1);
     }

   while(1)
     {
       nr = fread(buf,1,BFSZ,stdin);
       if(nr<=0)
	 {
	   fprintf(stderr,"EOF on stdin, shutting down audio ...\n");
	   audio_shutdown();
	   fprintf(stderr,"... done\n");
	   exit(0);
	 }
       
       n = 0;
       while(n<nr)
	 {
	   nw = audio_write(buf+n,nr-n,0);
	   if(nw<0)
	     {
	       fprintf(stderr,"audio_write error\n");
	       audio_shutdown();
	       fprintf(stderr,"... done\n");
	       exit(1);
	     }
	   n += nw;
	   if(n<nr)
	     {
	       fprintf(stderr,"audio_write: buffer full\n");
	       sleep(1);
	     }
	 }
     }
}
#endif



