/*

    Multifile recording version

	  Modified by: Mogi, Kazhiro <ee69011@isc.meiji.ac.jp>

 */

/*
    lavrec - Linux Audio Video RECord

    Copyright (C) 1999 Rainer Johanni <Rainer@Johanni.de>

    Capture motion video from the IOMEGA Buz to an AVI or Quicktime
    file together with capturing audio from the soundcard.

    Usage: lavrec [options] filename

    where options are as follows:

    -f [aq]      Format, 'a' for AVI (default) 'q' for quicktime

    -i [pPnNa]   Input Source:
                 'p' PAL  Composite Input
                 'P' PAL  SVHS-Input
                 'n' NTSC Composite Input
                 'N' NTSC SVHS-Input
                 'a' (or every other letter) Autosense (default)

    -d [124]      decimation, must be 1,2 or 4

    -q num        quality, must be beetween 0 and 100, default is 40

    -t num        Time to capture in seconds, default is unlimited
                 (use ^C to stop capture!)

   *** Audio settings ***

    -a num        Audio size in bits, must be 0 (no audio), 8 or 16 (default)

    -r num        Audio rate, must be a permitted sampling rate for your soundcard
                 default is 44100.

    -s num        stereo, must be 0 (mono) or 1 (stereo, default)

    -l num        Audio level to use for recording, must be beetwen 0 and 100
                  or -1 (for not touching the mixer settings at all), default is 85

   *** Special capture settings ***

    -n num        Number of MJPEG capture buffers (default 64)

    -b num        Buffersize of MJPEG buffers in KB (default 512)

	-o num        Odd/Even, must be 0 (Even first) or 1 (Odd first)

   *** Decoder color tone settings ***

    -B num        Brightness, must be 0 - 65535 (default 32768)

    -C num        Contrast, must be 0 - 65535 (default 32768)
	
    -H num        Hue, must be 0 - 65535 (default 32768)
	
    -S num        Saturation, must be 0 - 65535 (default 32768)

   *** Divide capture setting ***

    -N num        Saving file number (default 1)
	-Z num        File dividing size in MB (default 1024)


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <linux/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <sys/resource.h>

#include <signal.h>
#include <stdlib.h>

#include <linux/videodev.h>
#include <linux/soundcard.h>
#include <buz.h>

#include "avilib.h"

/* Set the default options here */

static char video_format = 'a';         /* 'a' for AVI, 'q' for Quicktime */
static char input_source = 'a';         /* 'a' for Auto detect, others see above */
static int  decimation   = 1;           /* decimation, 1, 2 or 4 */
static int  quality      = 40;          /* determines target size of JPEG code,
                                           100=max. 1/2 CompRatio */
static int  record_time  = 100000;      /* "unlimited" */

static int audio_size    = 16;          /* size of an audio sample: 8 or 16 bits,
                                           0 for no audio */
static int audio_rate    = 44100;       /* sampling rate for audio */
static int stereo        = 1;           /* 0: capture mono, 1: capture stereo */
static int audio_lev     = 85;         /* Level of Audio input,
                                            0..100: Recording level
                                           -1:  don't change mixer settings */
static int MJPG_nbufs    = 64;          /* Number of MJPEG buffers */
static int MJPG_bufsize  = 512;         /* Size   of MJPEG buffers */

static int input_brightness = 32768;   /* Brightness */
static int input_contrast   = 32768;   /* Contrast   */
static int input_hue        = 32768;   /* Hue        */
static int input_saturation = 32768;   /* Saturation */

static int output_file_num  = 1;       /* number of output file */
static int output_file_size = 1024;    /* size of file dividing */

static int MJPG_odd_even = 1;          /* Odd First */


/* On some systems MAP_FAILED seems to be missing */

#ifndef MAP_FAILED
#define MAP_FAILED ( (caddr_t) -1 )
#endif

#ifdef QUICKTIME4LINUX

#include <quicktime.h>
static quicktime_t *qt_file;

#endif

static int width, height;
static int input, norm;

static int mixer_set = 0;
static int mixer_volume_saved = 0;
static int mixer_recsrc_saved = 0;
static int mixer_linelv_saved = 0;

/*
   Set the sound mixer:
   flag = 1 : set for recording from the line input
   flag = 0 : restore previously saved values
*/

void set_mixer(int flag)
{
   int fd, recsrc, level, status, numerr;

   /* Avoid restoring anything when nothing was set */

   if (flag==0 && mixer_set==0) return;

   fd = open("/dev/mixer", O_RDONLY);
   if (fd == -1)
   {
      fprintf(stderr,"Unable to open sound mixer device /dev/mixer\n");
      fprintf(stderr,"Try setting the sound mixer with another tool!!!\n");
      return;
   }

   mixer_set = 1;

   if(flag)
   {
      /* Save the values we are going to change */

      numerr = 0;
      status = ioctl(fd, SOUND_MIXER_READ_VOLUME, &mixer_volume_saved);
      if (status == -1) numerr++;
      status = ioctl(fd, SOUND_MIXER_READ_RECSRC, &mixer_recsrc_saved);
      if (status == -1) numerr++;
      status = ioctl(fd, SOUND_MIXER_READ_LINE  , &mixer_linelv_saved);
      if (status == -1) numerr++;
      if (numerr) 
      {
         fprintf(stderr,"Unable to save sound mixer settings\n"
                        "Restore your favorite setting with another tool\n"
                        "after capturing finished\n");
         mixer_set = 0; /* Avoid restoring the wrong values */
      }

      /* Set the recording source to the line input, 
         the level of the line input to audio_lev,
         the output volume to zero (to avoid audio feedback
         when using a camera build in microphone */

      numerr = 0;

      recsrc = SOUND_MASK_LINE;
      status = ioctl(fd, SOUND_MIXER_WRITE_RECSRC, &recsrc);
      if (status == -1) numerr++;

      level = 256*audio_lev + audio_lev; /* left and right channel */
      status = ioctl(fd, SOUND_MIXER_WRITE_LINE, &level);
      if (status == -1) numerr++;

      level = 0;
      status = ioctl(fd, SOUND_MIXER_WRITE_VOLUME, &level);
      if (status == -1) numerr++;

      if (numerr) 
      {
         fprintf(stderr,"Unable to set the sound mixer correctly\n"
                        "Audio capture might not be successfull\n"
                        "Try setting the sound mixer with another tool!!!\n");
      }
   }
   else
   {
      /* Restore previously saved settings */

      numerr = 0;
      status = ioctl(fd, SOUND_MIXER_WRITE_RECSRC, &mixer_recsrc_saved);
      if (status == -1) numerr++;
      status = ioctl(fd, SOUND_MIXER_WRITE_LINE,   &mixer_linelv_saved);
      if (status == -1) numerr++;
      status = ioctl(fd, SOUND_MIXER_WRITE_VOLUME, &mixer_volume_saved);
      if (status == -1) numerr++;

      if (numerr) 
      {
         fprintf(stderr,"Unable to restore sound mixer settings\n"
                        "Restore your favorite setting with another tool\n");
      }
   }

   close(fd);

}

/* Shut down the audio task and rest the mixers.
   CleanUpAudio is nly called indirectly through atexit()
*/

void CleanUpAudio(void)
{
   if(audio_size) audio_shutdown();
   set_mixer(0);
}

/* The signal handler only sets a flag which makes the main program
   to exit the video recording loop.
   This way we avoid race conditions when ^C is pressed during
   writing to the AVI File
*/

static int VideoExitFlag = 0;

void SigHandler(int sig_num)
{
   VideoExitFlag = 1;
}

void Usage(char *progname)
{
   fprintf(stderr, "Usage: %s [options] <filename>\n", progname);
   fprintf(stderr, "where options are:\n");
   fprintf(stderr, "   -f [aq]    Format AVI/Quicktime\n");
   fprintf(stderr, "   -i [pPnNa] Input Source\n");
   fprintf(stderr, "   -d [124]   Decimation\n");
   fprintf(stderr, "   -q num     Quality [%%]\n");
   fprintf(stderr, "   -t num     Capture time\n");
   fprintf(stderr, "   -a num     Audio size, 0 for no audio, 8 or 16\n");
   fprintf(stderr, "   -r num     Audio rate [Hz]\n");
   fprintf(stderr, "   -s num     Stereo, 0 for mono, 1 for stereo\n");
   fprintf(stderr, "   -l num     Recording level [%%], -1 for mixers not touched\n");
   fprintf(stderr, "   -n num     # of MJPEG buffers\n");
   fprintf(stderr, "   -b num     Size of MJPEG buffers [Kb]\n");
   fprintf(stderr, "   -o num     Odd/Even, 0 for even first, 1 for odd first\n");
   fprintf(stderr, "   -B num     Brightness\n");
   fprintf(stderr, "   -C num     Contrast\n");
   fprintf(stderr, "   -H num     Hue\n");
   fprintf(stderr, "   -S num     Saturation\n");
   exit(1);
}

void cut_sfx(char *dest, char *src)
{
	int len;
	int i;

	len = strlen(src);
	
	for(i=len-1;i>=0;i--){
		if(src[i] == '.'){
			len = i;
			break;
		}
	}

	strncpy(dest, src, len);
	dest[len] = '\0';
}


main(int argc, char ** argv)
{
   int video_dev;
   int res;
   int fc;
   int i, n, nerr;
   int fnlen;
   int nframes;
   int count;
   int first_lost;
   off_t pos;
   struct timeval first_time;
   double time;
   char * MJPG_buff;
   char AUDIO_buff[65536];
   AVI_file *af;
   char **outfiles;
   char basename[FILENAME_MAX];
   struct buz_status bstat;
   struct buz_params bparm;
   struct video_picture bpict;
   struct buz_requestbuffers breq;
   struct buz_sync bsync;

   /* check usage */
   if (argc < 2)  Usage(argv[0]);

   /* Get options */

   nerr = 0;

   while( (n=getopt(argc,argv,"f:i:d:q:t:a:r:s:l:n:o:b:B:C:H:S:N:Z:")) != EOF){
   fflush(stdout);
      switch(n) {
	  case 'f':
		  video_format = optarg[0];
		  if(optarg[0]!='a' && optarg[0]!='q'){
			  fprintf(stderr,"Format (-f option) must be a or q\n");
			  nerr++;
		  }
		  break;
		  
	  case 'i':
		  input_source = optarg[0];
		  break;
		  
	  case 'd':
		  decimation = atoi(optarg);
		  if(decimation != 1 && decimation != 2 && decimation != 4){
			  fprintf(stderr,"decimation = %d invalid (must be 1, 2 or 4)\n",
                              decimation);
			  nerr++;
		  }
		  break;

	  case 'q':
		  quality = atoi(optarg);
		  if(quality<0 || quality>100){
			  fprintf(stderr,"quality = %d invalid (must be 0 ... 100)\n",quality);
			  nerr++;
		  }
		  break;

	  case 't':
		  record_time = atoi(optarg);
		  if(record_time<=0){
			  fprintf(stderr,"record_time = %d invalid\n",record_time);
			  nerr++;
		  }
		  break;

	  case 'a':
		  audio_size = atoi(optarg);
		  if(audio_size != 0 && audio_size != 8 && audio_size != 16){
			  fprintf(stderr,"audio_size = %d invalid (must be 0, 8 or 16)\n",
                              audio_size);
			  nerr++;
		  }
		  break;

	  case 'r':
		  audio_rate = atoi(optarg);
		  if(audio_rate<=0){
			  fprintf(stderr,"audio_rate = %d invalid\n",audio_rate);
			  nerr++;
		  }
		  break;

	  case 's':
		  stereo = atoi(optarg);
		  if(stereo != 0 && stereo != 1){
			  fprintf(stderr,"stereo = %d invalid (must be 0 or 1)\n",stereo);
			  nerr++;
		  }
		  break;

	  case 'l':
		  audio_lev = atoi(optarg);
		  if(audio_lev<-1 || audio_lev>100){
			  fprintf(stderr,"recording level = %d invalid (must be 0 ... 100 or -1)\n",
					  audio_lev);
			  nerr++;
		  }
		  break;

	  case 'n':
		  MJPG_nbufs = atoi(optarg);
		  break;

	  case 'b':
		  MJPG_bufsize = atoi(optarg);
		  break;

	  case 'o':
		  MJPG_odd_even = atoi(optarg);
		  if(MJPG_odd_even != 0 && MJPG_odd_even != 1){
			  fprintf(stderr, "Odd/Even = %d invalid (must be 0 or 1)\n",MJPG_odd_even);
			  nerr++;
		  }
		  break;

	  case 'B':
		  input_brightness = atoi(optarg);
		  if(input_brightness < 0 && input_brightness > 65535){
			  fprintf(stderr, "Brightness = %d invalid (must be 0 to 65535)\n",input_brightness);
			  nerr++;
		  }
		  break;

	  case 'C':
		  input_contrast = atoi(optarg);
		  if(input_contrast < 0 && input_contrast > 65535){
			  fprintf(stderr, "Contrast = %d invalid (must be 0 to 65535)\n",input_contrast);
			  nerr++;
		  }
		  break;

	  case 'H':
		  input_hue = atoi(optarg);
		  if(input_hue < 0 && input_hue > 65535){
			  fprintf(stderr, "Hue = %d invalid (must be 0 to 65535)\n",input_hue);
			  nerr++;
		  }
		  break;
		  
	  case 'S':
		  input_saturation = atoi(optarg);
		  if(input_saturation < 0 && input_saturation > 65535){
			  fprintf(stderr, "Saturation = %d invalid (must be 0 to 65535)\n",input_brightness);
			  nerr++;
		  }
		  break;

	  case 'N':
		  output_file_num = atoi(optarg);
		  if(output_file_num > FOPEN_MAX){
			  fprintf(stderr, "File number = %d invalid (must be under %d)\n",output_file_num, FOPEN_MAX);
			  nerr++;
		  }
		  break;

	  case 'Z':
		  output_file_size = atoi(optarg);
		  if(output_file_size > 2000){
			  fprintf(stderr, "Filesize = %d MB invalid (must be under 2000)\n",output_file_size);
			  nerr++;
		  }
		  break;
		  
	  default:
            nerr++;
            break;
      }
   }

   if(optind>=argc) nerr++;

   if(nerr) Usage(argv[0]);

   printf("\nRecording parameters:\n\n");
   printf("Output format:      %s\n",video_format=='a'?"AVI":"Quicktime");
   printf("Input Source:       ");
   switch(input_source)
   {
      case 'p': printf("Composite PAL\n"); break;
      case 'P': printf("S-VHS PAL\n"); break;
      case 'n': printf("Composite NTSC\n"); break;
      case 'N': printf("S-VHS NTSC\n"); break;
      default:  printf("Auto detect\n");
   }
   printf("Decimation:         %d\n",decimation);
   printf("Quality:            %d\n",quality);
   printf("Odd/Even:           %s\n",MJPG_odd_even ? "Odd first":"Even first");
   printf("Recording time:     %d sec\n",record_time);

   printf("\n");
   printf("Brightness:         %d\n",input_brightness);
   printf("Contrast:           %d\n",input_contrast);
   printf("Saturation:         %d\n",input_saturation);
   printf("Hue:                %d\n",input_hue);

   printf("\n");
   printf("File dividing size: %d MB\n",output_file_size);
   output_file_size = output_file_size * 1024 * 1024;
   printf("# of saving file:   %d\n",output_file_num);
   
   if(audio_size)
   {
      printf("\nAudio parameters:\n\n");
      printf("Audio sample size:           %d bit\n",audio_size);
      printf("Audio sampling rate:         %d Hz\n",audio_rate);
      printf("Audio is %s\n",stereo ? "STEREO" : "MONO");
      if(audio_lev!=-1)
         printf("Audio input recording level: %d %%\n",audio_lev);
      else
         printf("Audio input recording level: Use mixer setting\n");
   }
   else
      printf("\nAudio disabled\n\n");

#ifndef QUICKTIME4LINUX

   if(video_format=='q')
   {
      fprintf(stderr,"Error: Quicktime support not compiled in\n");
      exit(1);
   }

#else

   /* At the moment, decimation == 1 not supported for Quicktime */

   if(video_format=='q' && decimation==1)
   {
      fprintf(stderr,"Error: Decimation==1 not supported for Quicktime in\n");
      exit(1);
   }

#endif

   /* Flush the Linux File buffers to disk */

   sync();

   /* set the sound mixer */

   if(audio_size && audio_lev>=0) set_mixer(1);

   /* Initialize the audio system if audio is wanted.
      This involves a fork of the audio task and is done before
      the video device and the output file is opened */

   if (audio_size)
   {
      res = audio_init(1,stereo,audio_size,audio_rate);
      if(res)
      {
         fprintf(stderr,"Error initializing Audio\n");
         exit(1);
      }
   }

   /* The audio system needs a exit processing (audio_shutdown()),
      the mixers also should be reset at exit.
      Since the AVI library just exits on errors, we register
      a function for the exit processing */

   atexit(CleanUpAudio);

   /* open the video device */

   video_dev = open("/dev/video", O_RDONLY);
   if (video_dev < 0) {
      perror("/dev/video");
      exit(1);
   }

   /* Set input and norm according to input_source,
      do an auto detect if neccessary */

   switch(input_source)
   {
      case 'p': input = 0; norm = 0; break;
      case 'P': input = 1; norm = 0; break;
      case 'n': input = 0; norm = 1; break;
      case 'N': input = 1; norm = 1; break;
      default:
         n = 0;
         printf("Auto detecting input and norm ...\n");
         for(i=0;i<2;i++)
         {
            printf("Trying %s ...\n",i==0?"Composite":"S-VHS");
            bstat.input = i;
            res = ioctl(video_dev,BUZIOC_G_STATUS,&bstat);
            if(res)
            {
               perror("BUZIOC_G_STATUS");
               fprintf(stderr,"Unable to get input status ... exiting\n");
               exit(1);
            }
            if(bstat.signal)
            {
               printf("input present: %s %s\n",bstat.norm?"NTSC":"PAL",
                                               bstat.color?"color":"no color");
               input = i;
               norm = bstat.norm;
               n++;
            }
            else
               printf("NO signal\n");
         }
         if(n==0)
         {
            fprintf(stderr,"Error: No input signal ... exiting\n");
            exit(1);
         }
         if(n==2)
         {
            fprintf(stderr,"Error: Input signal on Composite AND S-VHS ... exiting\n");
            exit(1);
         }
   }

   /* Query and set params for capture */

   res = ioctl(video_dev, BUZIOC_G_PARAMS, &bparm);
   if (res < 0) { perror("BUZIOC_G_PARAMS"); exit(1); }

   bparm.input      = input;
   bparm.norm       = norm;
   bparm.decimation = decimation;
   bparm.quality    = quality;
   bparm.odd_even   = MJPG_odd_even;

   if(video_format=='a')
   {
      /* The APP0 marker is evaluated by some Windows programs for AVIs*/

      bparm.APPn    = 0;
      bparm.APP_len = 14;
      for(n=0;n<14;n++) bparm.APP_data[n] = 0;
      bparm.APP_data[0] = 'A';
      bparm.APP_data[1] = 'V';
      bparm.APP_data[2] = 'I';
      bparm.APP_data[3] = '1';

      if(bparm.decimation==1)
      {
		  bparm.APP_data[4] = bparm.odd_even ? 1 : 2;
      }
   }
   else
   {
      /* Quicktime uses the APP1 marker for MJPA, since we don't support
         MJPA right now, suppress APP markers at all */

      bparm.APP_len = 0;
   }

   res = ioctl(video_dev, BUZIOC_S_PARAMS, &bparm);
   if (res < 0) { perror("BUZIOC_S_PARAMS"); exit(1); }

   
   /* Set decorder picture */
   
   bpict.brightness = input_brightness;
   bpict.contrast = input_contrast;
   bpict.colour = input_saturation;
   bpict.hue = input_hue;
   
   res = ioctl(video_dev, VIDIOCSPICT, &bpict);
   if( res < 0) { perror("VIDIOCSPICT"); exit(1); }

   /* Open output file */

   if(video_format=='a'){
	   if(output_file_num > 1){
		   af = calloc(output_file_num, sizeof(AVI_file));
		   outfiles = malloc(output_file_num * sizeof(char *));
		   cut_sfx(basename, argv[optind]);
		   fnlen = strlen(basename) + 8;
		   for(fc=0;fc<output_file_num;fc++){
			   outfiles[fc] = malloc(sizeof(char)*fnlen);
			   sprintf(outfiles[fc], "%s%03d%s",basename, fc, ".avi");
			   af[fc] = AVI_open_output_file(outfiles[fc]);
			   AVI_start_list(&af[fc], "movi");
		   }
	   }else{
		   af = calloc(1, sizeof(AVI_file));
		   outfiles = malloc(sizeof(char *));
		   fnlen = strlen(argv[optind]) + 1;
		   outfiles[0] = malloc(fnlen);
		   strcpy(outfiles[0], argv[optind]);
		   af[0] = AVI_open_output_file(outfiles[0]);
		   AVI_start_list(&af[0], "movi");
	   }
   }
#ifdef QUICKTIME4LINUX
   else
   {
      /* open quicktime output file */

      /* since the documentation says that the file should be empty,
         we try to remove it first */

      res = remove(argv[optind]);

      qt_file = quicktime_open(argv[optind], 1, 1);
      if(qt_file == 0)
      {
         fprintf(stderr,"Can not open quicktime file %s\n",argv[2]);
         exit(1);
      }
      if (audio_size) quicktime_set_audio(qt_file, (stereo ? 2 : 1), audio_rate,
                                          audio_size, QUICKTIME_RAW);
   }
#endif

   printf("Image size will be %dx%d, %d field(s) per buffer\n",
          bparm.img_width/bparm.HorDcm, bparm.img_height/bparm.VerDcm, bparm.field_per_buff);

   width  = bparm.img_width/bparm.HorDcm;
   height = bparm.img_height/bparm.VerDcm*bparm.field_per_buff;

#ifdef QUICKTIME4LINUX
   if(video_format=='q')
   {
      quicktime_set_video(qt_file, 1, width, height,
        (norm == VIDEO_MODE_PAL ? 25.0 : 29.97),QUICKTIME_JPEG);
   }
   /* For AVI, width and height are set in AVI_write_header(),
      which is called when all is finished! */
#endif

   /* Request buffers */

   breq.count = MJPG_nbufs;
   breq.size  = MJPG_bufsize*1024;
   res = ioctl(video_dev, BUZIOC_REQBUFS,&breq);
   if (res < 0) { perror("BUZIOC_REQBUFS"); exit(1); }
   printf("Got %d buffers of size %d KB\n",breq.count,breq.size/1024);

   /* Map the buffers */
   MJPG_buff = mmap(0, breq.count*breq.size, PROT_READ, MAP_SHARED, video_dev, 0);
   if (MJPG_buff == MAP_FAILED) { perror("mmap"); exit(1); }

   /* Assure proper exit handling if the user presses ^C during recording */

   signal(SIGINT,SigHandler);

   /* Synchronize with audio (hopefully) */

   if (audio_size) audio_start();

   /* Queue all buffers, this also starts streaming capture */
   
   for(n=0;n<breq.count;n++){
	   res = ioctl(video_dev, BUZIOC_QBUF_CAPT, &n);
      if (res < 0) { perror("BUZIOC_QBUF_CAPT"); exit(1); }
   }

   /* The video capture loop */

   fc = 0;
   printf("Filename: %s\n", outfiles[fc]);
   nframes = 0;
   count = (norm==0) ? record_time*25 : record_time*30;
   
   while (nframes < count && !VideoExitFlag){

	   /* Filesize check */
	   if(lseek(af[fc].fd,0,SEEK_CUR) > output_file_size){
		   fc++;
		   if(fc >= output_file_num){
			   break;
		   }
		   printf("\nFilename: %s\n", outfiles[fc]);
	   }

	   /* sync on a frame */

	   res = ioctl(video_dev, BUZIOC_SYNC, &bsync);
	   if (res < 0) { perror("BUZIOC_SYNC"); break; }

	   nframes++;

	   if(nframes==1){
		   first_time = bsync.timestamp;
		   first_lost = bsync.seq-nframes;
	   }
	   
	   
	   time = bsync.timestamp.tv_sec - first_time.tv_sec
           + 1.e-6*(bsync.timestamp.tv_usec - first_time.tv_usec);
	   
	   printf("\rframe %5d, length=%6u seq=%5u (lost: %4d) time=%8.3f",
			  nframes - 1, bsync.length, bsync.seq,
			  bsync.seq-nframes-first_lost, time);
	   fflush(stdout);
	   
	   /* write it out */
	   
	   if(video_format=='a'){
		   AVI_add_frame(&af[fc], MJPG_buff + bsync.frame*breq.size, bsync.length);
	   }
#ifdef QUICKTIME4LINUX
	   else{
		   quicktime_write_frame(qt_file, MJPG_buff + bsync.frame*breq.size, bsync.length, 0);
	   }
#endif

	   /* Re-queue the buffer */

	   res = ioctl(video_dev, BUZIOC_QBUF_CAPT, &bsync.frame);
	   if (res < 0) { perror("BUZIOC_QBUF_CAPT"); break; }


	   /* Care about audio */

	   if (!audio_size) continue;
	   
	   /* Quicktime needs Big Endian Audio, AVI Little Endian */

	   res = audio_read(AUDIO_buff,sizeof(AUDIO_buff),(video_format=='a' ? 0 : 1));

	   if(res==0) continue; /* No audio sample present */

	   if(res<0)  continue; /* Some error, we ignore it */

	   /* Got an audio sample, write it out */

	   if(video_format=='a'){
		   AVI_add_audio(&af[fc], AUDIO_buff,res);
	   }
#ifdef QUICKTIME4LINUX
	   else{
		   /* Calculate the number of audio samples */
		   n = res;
		   if(stereo) n = n/2;
		   if(audio_size==16) n = n/2;

		   quicktime_write_audio(qt_file,AUDIO_buff,n,0);
	   }
#endif
   }

   /* Close the video output file */

   if(video_format=='a'){
	   /* Finish movi list */
	   
	   for(fc=0;fc<output_file_num;fc++){
		   
		   /* Filesize check */
		   if(lseek(af[fc].fd,0,SEEK_CUR) == DATASTART){
			   
			   /* disused files */
			   close(af[fc].fd);
			   unlink(outfiles[fc]);
			   
		   }else{

			   /* used files */
			   AVI_finish_list(&af[fc],"movi");

			   AVI_write_header(&af[fc],width,height,norm,(audio_size ? 1 : 0),stereo,audio_size,audio_rate);
			   
		   }

		   free(outfiles[fc]);
	   }
	   free(af);
	   free(outfiles);
   }
#ifdef QUICKTIME4LINUX
   else
   {
      quicktime_close(qt_file);
   }
#endif

   /* Audio and mixer exit processing is done with atexit() */

   fprintf(stderr,"\n\nClean exit ...\n");
   exit(0);
}
